import shutil
import textwrap

from pygments import highlight

# I do not understand this, but pylint (2.17.4) raises
# aissist/code_formatter.py:5:0: E0611: No name 'TerminalTrueColorFormatter' in module 'pygments.formatters' (no-name-in-module)
# Nevertheless, the code works fine so I am ignoring it.
from pygments.formatters import TerminalTrueColorFormatter  # pylint: disable=E0611
from pygments.lexers import guess_lexer


class CodeFormatter:
    def __init__(self, color_scheme: str) -> None:
        self.color_scheme = color_scheme

    def bold_single_backticks(self, text: str) -> str:
        """Formats a given string by making text enclosed in single backticks bold,
        using ANSI escape codes.
        """
        in_backticks = False
        result = ""
        for c in text:
            if c == "`":
                if in_backticks:
                    result += "\033[0m"  # Reset text formatting
                    in_backticks = False
                else:
                    result += "\033[1m"  # Bold text
                    in_backticks = True
            else:
                result += c

        if in_backticks:
            result += "\033[0m"  # Reset text formatting

        return result

    def highlight_codeblocks(self, markdown: str) -> None:
        """Highlights code blocks in a markdown string and prints them to the console
        using pygments library.
        """

        terminal_size = shutil.get_terminal_size()
        terminal_width = terminal_size.columns
        inside_block = False
        current_block = ""
        for line in markdown.split("\n"):
            if line.startswith("```"):
                if inside_block is True:
                    # We are exiting a block, print it
                    lexer = guess_lexer(current_block.rstrip("\n"))
                    formatted_code = highlight(
                        current_block,
                        lexer,
                        TerminalTrueColorFormatter(style=self.color_scheme),
                    )
                    print(formatted_code)
                else:
                    # We are entering a block
                    current_block = ""
                inside_block = not inside_block

            elif inside_block is True:
                current_block += line + "\n"
            else:
                line = self.bold_single_backticks(line)
                line = textwrap.fill(line, terminal_width - 1)
                print(line)



import os
import sys

import openai
from prompt_toolkit import PromptSession
from prompt_toolkit.styles import Style

from .config import Config
from .exceptions import AIssistError
from .model import Model, OpenAIMessage
from .spinner import Spinner
from .version import __version__

try:
    openai.api_key = os.environ["OPENAI_API_KEY"]
except KeyError:
    print("Please put your API key in the OPENAI_API_KEY environment variable.")
    sys.exit(1)


def prompt_continuation(
    width: int, line_number: int, is_soft_wrap: int  # pylint: disable=W0613
) -> str:
    """When the user is typing a multi-line prompt, this function is called to
    determine what the prompt should look like on the next line.

    n.b. there are unused parameters to this function, but they are required by
    the PromptSession.prompt_continuation interface."""
    return "." * (width - 1) + " "


def print_streaming_message(
    model: Model, messages: list[OpenAIMessage], config: Config
) -> OpenAIMessage:
    """Prints a message that is being streamed from the API"""

    new_message_str = ""
    for printable_chunk in model.stream_call(messages, config):
        sys.stdout.write(printable_chunk)
        sys.stdout.flush()
        new_message_str += printable_chunk
    print("\n")
    new_message: OpenAIMessage = {"role": "assistant", "content": new_message_str}
    return new_message


def print_message(
    model: Model, messages: list[OpenAIMessage], config: Config
) -> OpenAIMessage:
    """Prints a message that is being returned from the API"""

    spinner = Spinner()
    spinner.start()

    new_message = model.call(messages, config)
    spinner.stop()
    print(" ")
    config.code_formatter.highlight_codeblocks(new_message["content"])
    print("\n")

    return new_message


def loop(config: Config, model: Model) -> None:
    """Main loop for the program"""

    # Move the fdollowing two lines into config, too
    style = Style.from_dict({"prompt": "#aaaaaa"})
    session: PromptSession = PromptSession(style=style)

    messages: list[OpenAIMessage] = [{"role": "system", "content": config.prompt}]

    get_completion_function = (
        print_message if config.get("no-stream") is True else print_streaming_message
    )

    while True:
        result = session.prompt(
            ">>> ", multiline=True, prompt_continuation=prompt_continuation
        )
        print()

        messages.append({"role": "user", "content": result})

        new_message = get_completion_function(model, messages, config)

        messages.append(new_message)


def main() -> None:
    print(f"AIssist v.{__version__}. ESCAPE followed by ENTER to send. Ctrl-C to quit")
    print("\n")

    config = Config()
    model = Model(config.get("model"))

    try:
        while True:
            loop(config, model)
    except (KeyboardInterrupt, EOFError):
        # Ctrl-C and Ctrl-D
        sys.exit(0)
    except AIssistError as e:
        print(e)
        sys.exit(1)
    except openai.error.OpenAIError as e:
        print(f"Received Error from openai: {str(e)}")
        sys.exit(1)
    except Exception:  # pylint: disable=W0706
        # If we didn't catch the error with one of the specific exceptions above,
        # print the stack trace and exit with a non-zero exit code.
        raise


if __name__ == "__main__":
    main()



DEFAULT_PROMPTS = {
    "code": (
        "A terse but helpful chatbot for coding questions",
        "You are a helpful but succinct chatbot that assists an impatient programmer by directly answering questions. You understand that the programmer is advanced enough to understand succinct phrases and prefers direct answers with little boilerplate.",
    )
}



"""Contains the Model class, which represents an OpenAI model."""
from typing import Generator, TypedDict, cast

import openai
import tiktoken
from retry import retry

from .config import Config
from .exceptions import InvalidParameterError


class OpenAIMessage(TypedDict):
    """Structure of an element in a conversation with the OpenAI API."""

    role: str
    content: str


class Model:
    # Map model to context length.
    model_lookup: dict[str, int] = {
        "gpt-3.5-turbo": 4096,
        "gpt-3.5-turbo-16k": 16384,
        "gpt-3.5-0613": 4096,
        "gpt-3.5-16k-0613": 16384,
        "gpt-4": 8192,
        "gpt-4-0613": 8192,
        "gpt-4-32k": 32768,
        "gpt-4-32k-0613": 32768,
    }

    def __init__(self, name: str):
        self.name = name

        try:
            self.context = self.model_lookup[self.name]
        except KeyError as exc:
            raise InvalidParameterError(f"Invalid model name: {self.name}") from exc

    @property
    def encoding(self) -> tiktoken.Encoding:
        return tiktoken.encoding_for_model(self.name)

    @retry(
        (
            openai.error.APIError,
            openai.error.APIConnectionError,
            openai.error.RateLimitError,
            openai.error.ServiceUnavailableError,
        ),
        tries=3,
        delay=2,
        backoff=2,
    )
    def call(self, messages: list[OpenAIMessage], config: Config) -> OpenAIMessage:
        self.trim_messages_to_context(messages, config.get("max_tokens"))

        # n.b. ignoring types because
        # error: Call to untyped function "create" of "ChatCompletion" in typed context
        response = openai.ChatCompletion.create(  # type: ignore
            model=self.name,
            temperature=config.get("temperature"),
            max_tokens=config.get("max_tokens"),
            messages=messages,
            stream=False,
        )

        message = response["choices"][0]["message"]
        return cast(OpenAIMessage, message)

    @retry(
        (
            openai.error.APIError,
            openai.error.APIConnectionError,
            openai.error.RateLimitError,
            openai.error.ServiceUnavailableError,
        ),
        tries=3,
        delay=2,
        backoff=2,
    )
    def stream_call(
        self, messages: list[OpenAIMessage], config: Config
    ) -> Generator[str, None, str]:
        self.trim_messages_to_context(messages, config.get("max_tokens"))

        # n.b. ignoring types because
        # error: Call to untyped function "create" of "ChatCompletion" in typed context
        response = openai.ChatCompletion.create(  # type: ignore
            model=self.name,
            temperature=config.get("temperature"),
            max_tokens=config.get("max_tokens"),
            messages=messages,
            stream=True,
        )

        buffer = ""
        for chunk in response:
            if chunk.choices[0]["finish_reason"] is not None:
                return buffer

            delta = chunk.choices[0]["delta"]["content"]
            yield delta

        return ""

    def messages_to_tokens(
        self, messages: list[OpenAIMessage], model: str = "gpt-3.5-turbo-0613"
    ) -> int:
        """Return the number of tokens used by a list of messages.

        https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
        """

        if model in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
        }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif model == "gpt-3.5-turbo-0301":
            tokens_per_message = (
                4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            )
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif "gpt-3.5-turbo" in model:
            return self.messages_to_tokens(messages, model="gpt-3.5-turbo-0613")
        elif "gpt-4" in model:
            return self.messages_to_tokens(messages, model="gpt-4-0613")
        else:
            raise NotImplementedError(
                f"""messages_to_tokens() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
            )
        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                num_tokens += len(self.encoding.encode(cast(str, value)))
                if key == "name":
                    num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    def trim_messages_to_context(
        self, messages: list[OpenAIMessage], max_tokens: int
    ) -> list[OpenAIMessage]:
        """Given a list of messages, discard old messages until the total number of
        tokens in the messages plus the max tokens in the completion is less than the
        context length of the model.
        """

        message_tokens = self.messages_to_tokens(messages, self.name)

        total_tokens = message_tokens + max_tokens

        if total_tokens > self.context:
            messages.pop(0)
            return self.trim_messages_to_context(messages, max_tokens)

        return messages



"""Configuation class for aissist."""

import argparse
import dataclasses
import os
from typing import Any, Type, TypeVar, cast

from tomlkit import comment, document, nl, parse, table

from .code_formatter import CodeFormatter
from .prompts import DEFAULT_PROMPTS

T = TypeVar("T")


@dataclasses.dataclass
class Parameter:
    type: Type
    value: Any
    comment: str = ""

    def __str__(self) -> str:
        return f"{self.value}"

    def set(self, value: T) -> None:
        if not isinstance(value, self.type):
            raise TypeError(f"Expected {self.type}, got {type(value)}")
        self.value = value


class Config:
    # The parameters known to AIssist.
    default_parameters: dict[str, Parameter] = {
        "model": Parameter(
            str, "gpt-3.5-turbo", comment="The default OpenAI model name to use"
        ),
        "prompt": Parameter(
            str,
            "code",
            comment="The default prompt to use. See the prompts.py file for a list of available prompts.",
        ),
        "no-stream": Parameter(
            bool, True, comment="Don't stream the output from the API."
        ),
        "color-scheme": Parameter(
            str,
            "paraiso-dark",
            comment="The default color scheme to use for code highlighting. See the pygments documentation for a list of available color schemes.",
        ),
        "temperature": Parameter(
            float,
            0.0,
            comment="A number in the range 0 to 1. Higher values lead to more unpredictable but creative outputs.",
        ),
        "max_tokens": Parameter(
            int,
            1000,
            comment="The maximum number of tokens to generate. Requests can use up to the model maximum tokens shared between prompt and response.",
        ),
    }

    def __init__(self) -> None:
        self.prompts: dict[str, str] = {}

        self.config_file = os.path.join(os.path.expanduser("~"), ".aissistrc")

        # Load the default parameters into self.parameters
        self.parameters: dict[str, Parameter] = self.default_parameters.copy()

        # If the config file does not exist, write the
        # defaults there.
        if not os.path.exists(self.config_file):
            self.write_config()
        else:
            # The configuration file was already present, read from it and overwrite
            # any parameters
            self.read_config()

        # Finally, if there are any command line parameters, overwrite the config
        # with them.
        self.add_command_line_args()

    @property
    def code_formatter(self) -> CodeFormatter:
        return CodeFormatter(self.parameters["color-scheme"].value)

    @property
    def prompt(self) -> str:
        prompt = self.parameters["prompt"].value
        return self.prompts[prompt]

    def add_command_line_args(self) -> None:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
        for name, param in self.default_parameters.items():
            if param.type == bool:
                # For a bool we need to add a store true and store false
                parser.add_argument(
                    f"--{name}",
                    action="store_true",
                    default=False,
                    help=param.comment,
                )
            else:
                parser.add_argument(
                    f"--{name}",
                    type=param.type,
                    default=param.value,
                    metavar=f"<{name}>",
                    help=param.comment,
                )
        args = parser.parse_args()

        for name, param in self.parameters.items():
            param.set(getattr(args, name.replace("-", "_")))

    def write_config(self) -> None:
        """Write the default configuration to the config file."""

        doc = document()
        doc.add(comment("Configuration file for aissist."))

        parameters = table()
        for name, param in self.parameters.items():
            parameters.add(nl())
            if len(param.comment) > 0:
                parameters.add(comment(param.comment))
            parameters.add(name, param.value)
        doc.add("parameters", parameters)

        prompts = table()

        for prompt, (prompt_description, prompt_str) in DEFAULT_PROMPTS.items():
            prompts.add(nl())
            prompts.add(comment(prompt_description))
            prompts.add(prompt, prompt_str)
            self.prompts[prompt] = prompt_str

        doc.add("prompts", prompts)

        with open(self.config_file, "w", encoding="utf-8") as f:
            f.write(doc.as_string())

    def read_config(self) -> None:
        """Read the configuration from the config file."""

        with open(self.config_file, "r", encoding="utf-8") as f:
            doc = parse(f.read())

        for name, param in cast(dict, doc["parameters"]).items():
            if name not in self.parameters:
                raise ValueError(
                    f"Unknown parameter {name}. Allowed parameters are: {', '.join(self.parameters.keys())}"
                )
            self.parameters[name].set(param)

        for name, prompt in cast(dict, doc["prompts"]).items():
            self.prompts[name] = prompt

    def get(self, parameter_name: str) -> Any:
        return self.parameters[parameter_name].value



class AIssistError(Exception):
    """Base class for all exceptions raised by AIssist"""


class InvalidParameterError(AIssistError):
    """Raised when an invalid parameter is passed to a function"""



import pytest

from aissist.model import Model
from aissist.exceptions import InvalidParameterError

def test_instantiation__invalid_model():
    """Test that an InvalidParameterError is raised when an invalid model is passed
    to Model"""

    with pytest.raises(InvalidParameterError):
        Model("not-a-real-model")